# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2021 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import wave

from threading import Event, Thread
from time import time
from unittest.mock import Mock
from os.path import join, isdir, dirname
from os import makedirs

from ovos_plugin_manager.microphone import OVOSMicrophoneFactory
from ovos_plugin_manager.vad import OVOSVADFactory
from ovos_dinkum_listener.voice_loop.voice_loop import DinkumVoiceLoop
from ovos_dinkum_listener.voice_loop.hotwords import HotwordContainer
from ovos_config.config import Configuration
from ovos_utils.messagebus import FakeBus
from ovos_utils.log import LOG
from ovos_utils.xdg_utils import xdg_data_home
from ovos_utils.sound import play_wav
from ovos_bus_client.message import Message
from neon_utils.file_utils import decode_base64_string_to_file
from neon_iris.client import NeonAIClient


class MockTransformers(Mock):
    def transform(self, chunk):
        return chunk, dict()


class NeonVoiceClient(NeonAIClient):
    def __init__(self, bus=None):
        self.config = Configuration()
        NeonAIClient.__init__(self, self.config.get("MQ"))
        self.bus = bus or FakeBus()
        self._mic = OVOSMicrophoneFactory.create(self.config)
        self._mic.start()
        self._hotwords = HotwordContainer(self.bus)
        self._hotwords.load_hotword_engines()
        self._vad = OVOSVADFactory.create(self.config)

        self._voice_loop = DinkumVoiceLoop(mic=self._mic,
                                           hotwords=self._hotwords,
                                           stt=Mock(),
                                           fallback_stt=Mock(),
                                           vad=self._vad,
                                           transformers=MockTransformers(),
                                           stt_audio_callback=self.on_stt_audio,
                                           listenword_audio_callback=self.on_hotword_audio)
        self._voice_loop.start()
        self._voice_thread = None

        self._stt_audio_path = join(xdg_data_home(), "iris", "stt")
        self._tts_audio_path = join(xdg_data_home(), "iris", "tts")
        if not isdir(self._stt_audio_path):
            makedirs(self._stt_audio_path)
        if not isdir(self._tts_audio_path):
            makedirs(self._tts_audio_path)

        self._listening_sound = join(dirname(__file__), "res",
                                     "start_listening.wav")

        self.run()

    def run(self):
        self._voice_thread = Thread(target=self._voice_loop.run, daemon=True)
        self._voice_thread.start()

    def on_stt_audio(self, audio_bytes: bytes, context: dict):
        LOG.info(f"Got {len(audio_bytes)} bytes of audio")
        wav_path = join(self._stt_audio_path, f"{time()}.wav")
        with open(wav_path, "wb") as wav_io, \
                wave.open(wav_io, "wb") as wav_file:
            wav_file.setframerate(self._mic.sample_rate)
            wav_file.setsampwidth(self._mic.sample_width)
            wav_file.setnchannels(self._mic.sample_channels)
            wav_file.writeframes(audio_bytes)

        self.send_audio(wav_path)
        LOG.debug("Sent Audio to MQ")

    def on_hotword_audio(self, audio: bytes, context: dict):
        payload = context
        msg_type = "recognizer_loop:wakeword"
        play_wav(self._listening_sound)
        LOG.info(f"Emitting hotword event: {msg_type}")
        # emit ww event
        self.bus.emit(Message(msg_type, payload, context))

    def handle_klat_response(self, message: Message):
        responses = message.data.get('responses')
        for lang, data in responses.items():
            text = data.get('sentence')
            LOG.info(text)
            file_basename = f"{hash(text)}.wav"
            genders = data.get('genders', [])
            for gender in genders:
                audio_data = data["audio"].get(gender)
                audio_file = join(self._tts_audio_path, lang, gender,
                                  file_basename)
                try:
                    decode_base64_string_to_file(audio_data, audio_file)
                except FileExistsError:
                    pass
                play_wav(audio_file)

    def handle_complete_intent_failure(self, message: Message):
        LOG.info(f"{message.data}")

    def handle_api_response(self, message: Message):
        LOG.info(f"{message.data}")

    def handle_error_response(self, message: Message):
        LOG.error(f"Got error response: {message.data}")

    def clear_caches(self, message: Message):
        pass

    def clear_media(self, message: Message):
        pass

    def shutdown(self):
        self._voice_loop.stop()
        self._voice_thread.join(30)
        NeonAIClient.shutdown(self)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2021 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
from pprint import pformat

import click

from os.path import expanduser, isfile
from time import sleep
from click_default_group import DefaultGroup
from ovos_utils.log import LOG

from neon_iris.util import load_config_file
from neon_iris.version import __version__


def _print_config():
    from ovos_config.config import Configuration
    config = Configuration().get('MQ')
    mq_endpoint = f"{config.get('server')}:{config.get('port', 5672)}"
    click.echo(f"Connecting to {mq_endpoint}")


@click.group("iris", cls=DefaultGroup,
             no_args_is_help=True, invoke_without_command=True,
             help="Iris: Interactive Relay for Intelligence Systems.\n\n"
                  "See also: iris COMMAND --help")
@click.option("--version", "-v", is_flag=True, required=False,
              help="Print the current version")
def neon_iris_cli(version: bool = False):
    if version:
        click.echo(f"Iris version {__version__}")


@neon_iris_cli.command(help="Create an MQ client session")
@click.option('--mq_config', '-m',
              help="Path to MQ Config file")
@click.option('--user-config', '-u',
              help="Path to User Config file")
@click.option('--lang', '-l', default="en-us",
              help="Language to accept input in")
@click.option('--audio', '-a', is_flag=True, default=False,
              help="Flag to enable audio playback")
def start_client(mq_config, user_config, lang, audio):
    from neon_iris.client import CLIClient
    _print_config()
    if mq_config:
        from ovos_config.locations import find_user_config
        click.echo(f"WARNING: Configuration should me moved to: "
                   f"{find_user_config()}.")
        mq_config = load_config_file(expanduser(mq_config))
    else:
        from ovos_config.config import Configuration
        mq_config = Configuration().get("MQ")
    if user_config:
        user_config = load_config_file(expanduser(user_config))
    client = CLIClient(mq_config, user_config)
    LOG.init({"level": logging.WARNING})  # TODO: Debug flag?

    client.audio_enabled = audio
    click.echo("Enter '!{lang}' to change language\n"
               "Enter '!quit' to quit.\n"
               "Enter '!mute' or '!unmute' to change audio playback")
    try:
        while True:
            query = click.prompt("Query")
            if query.startswith('!'):
                if query == "!quit":
                    break
                elif query == "!mute":
                    click.echo("Disabling Audio Playback")
                    client.audio_enabled = False
                elif query == "!unmute":
                    click.echo("Enabling Audio Playback")
                    client.audio_enabled = True
                else:
                    query = query.lstrip('!')
                    query = expanduser(query)
                    if isfile(query):
                        client.send_audio(query, lang, client.username,
                                          client.user_profiles)
                    else:
                        lang = query.split()[0]
                        client.user_profiles[0]["speech"]["secondary_tts_language"] = lang
                        click.echo(f"Language set to {lang}")
            else:
                client.send_utterance(query, lang)
                # Pad prompt for multiple responses
                sleep(1)
    except Exception as e:
        click.echo(e)
    click.echo("Shutting Down Client")
    client.shutdown()


@neon_iris_cli.command(help="Create an MQ listener session")
def start_listener():
    from neon_iris.voice_client import NeonVoiceClient
    from ovos_utils import wait_for_exit_signal
    client = NeonVoiceClient()
    _print_config()
    wait_for_exit_signal()
    client.shutdown()


@neon_iris_cli.command(help="Create a GradIO Client session")
def start_gradio():
    from neon_iris.web_client import GradIOClient
    _print_config()
    try:
        chat = GradIOClient()
        chat.run()
    except OSError:
        click.echo("Unable to connect to MQ server")


@neon_iris_cli.command(help="Create a Web Voice Satellite session")
@click.option("--port", "-p", default=8000, help="Port to run on, defaults to 8000")
@click.option("--host", default="0.0.0.0", help="Host to run on, defaults to 0.0.0.0")
def start_websat(port, host):
    from neon_iris.web_sat_client import app
    _print_config()
    try:
        import uvicorn
        uvicorn.run(app, host=host, port=port)
    except OSError:
        click.echo("Unable to connect to MQ server")


@neon_iris_cli.command(help="Query Neon Core for supported languages")
def get_languages():
    from neon_iris.util import query_neon
    _print_config()
    resp = query_neon("neon.languages.get", {})
    click.echo(pformat(resp))


@neon_iris_cli.command(help="Transcribe an audio file")
@click.option('--lang', '-l', default='en-us',
              help="language of input audio")
@click.argument("audio_file")
def get_stt(audio_file, lang):
    from neon_iris.util import get_stt
    _print_config()
    resp = get_stt(audio_file, lang)
    click.echo(pformat(resp))


@neon_iris_cli.command(help="Transcribe an audio file")
@click.option('--lang', '-l', default='en-us',
              help="language of input audio")
@click.argument("utterance")
def get_tts(utterance, lang):
    from neon_iris.util import get_tts
    _print_config()
    resp = get_tts(utterance, lang)
    click.echo(pformat(resp))


# Backend
@neon_iris_cli.command(help="Query a weather endpoint")
@click.option('--unit', '-u', default='imperial',
              help="units to use ('metric' or 'imperial')")
@click.option('--latitude', '--lat', default=47.6815,
              help="location latitude")
@click.option('--longitude', '--lon', default=-122.2087,
              help="location latitude")
@click.option('--api', '-a', default='onecall',
              help="api to query ('onecall' or 'weather')")
def get_weather(unit, latitude, longitude, api):
    from neon_iris.util import query_api
    _print_config()
    query = {"lat": latitude,
             "lon": longitude,
             "units": unit,
             "api": api,
             "service": "open_weather_map"}
    resp = query_api(query)
    click.echo(pformat(resp))


@neon_iris_cli.command(help="Query a stock price endpoint")
@click.argument('symbol')
def get_stock_quote(symbol):
    from neon_iris.util import query_api
    _print_config()
    query = {"symbol": symbol,
             "api": "quote",
             "service": "alpha_vantage"}
    resp = query_api(query)
    click.echo(pformat(resp))


@neon_iris_cli.command(help="Query a stock symbol endpoint")
@click.argument('company')
def get_stock_symbol(company):
    from neon_iris.util import query_api
    _print_config()
    query = {"company": company,
             "api": "symbol",
             "service": "alpha_vantage"}
    resp = query_api(query)
    click.echo(pformat(resp))


@neon_iris_cli.command(help="Query a WolframAlpha endpoint")
@click.option('--api', '-a', default='short',
              help="Wolfram|Alpha API to query")
@click.option('--unit', '-u', default='imperial',
              help="units to use ('metric' or 'imperial')")
@click.option('--latitude', '--lat', default=47.6815,
              help="location latitude")
@click.option('--longitude', '--lon', default=-122.2087,
              help="location latitude")
@click.argument('question')
def get_wolfram_response(api, unit, latitude, longitude, question):
    from neon_iris.util import query_api
    _print_config()
    query = {"api": api,
             "units": unit,
             "latlong": f"{latitude},{longitude}",
             "query": question,
             "service": "wolfram_alpha"}
    resp = query_api(query)
    click.echo(pformat(resp))


@neon_iris_cli.command(help="Query Map Maker Geolocation endpoint")
@click.option('--latitude', '--lat',
              help="location latitude")
@click.option('--longitude', '--lon',
              help="location latitude")
@click.option('--query', '-q', help="Name of location to look up")
def get_mapmaker_response(latitude, longitude, query):
    from neon_iris.util import query_api
    _print_config()
    query = {"lat": latitude,
             "lon": longitude,
             "address": query,
             "service": "map_maker"}
    resp = query_api(query)
    click.echo(pformat(resp))


@neon_iris_cli.command(help="Converse with an LLM")
@click.option('--llm', default="chat_gpt",
              help="LLM Queue to interact with ('chat_gpt' or 'fastchat')")
def start_llm_chat(llm):
    from neon_iris.llm import LLMConversation
    _print_config()
    conversation = LLMConversation(llm)
    while True:
        query = click.prompt(">")
        resp = conversation.get_response(query)
        click.echo(resp)


@neon_iris_cli.command(help="Converse with an LLM")
def get_coupons():
    from neon_iris.util import get_brands_coupons
    data = get_brands_coupons()
    click.echo(pformat(data))


@neon_iris_cli.command(help="Parse a Neon CCL script")
@click.argument("script_file")
def parse_script(script_file):
    from neon_iris.util import parse_ccl_script
    data = parse_ccl_script(script_file)
    click.echo(pformat(data))

# TODO: email, metrics



"""Runs a web server that serves the Neon AI Web UI and Voice Satellite."""
# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2024 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
from os import makedirs
from os.path import isdir, join
from threading import Event
from time import time
from typing import Dict, Optional, Sequence
from uuid import uuid4

import numpy as np
import resampy
from fastapi import APIRouter, FastAPI, Request, WebSocket
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from neon_utils.file_utils import decode_base64_string_to_file
from openwakeword import Model
from ovos_bus_client import Message
from ovos_config import Configuration
from ovos_utils import LOG
from ovos_utils.xdg_utils import xdg_data_home

from neon_iris.client import NeonAIClient
from neon_iris.models.web_sat import UserInput, UserInputResponse


class WebSatNeonClient(NeonAIClient):
    """Neon AI Web UI and Voice Satellite client."""

    def __init__(self, lang: str = ""):
        config = Configuration()
        self.config = config.get("iris") or dict()
        self.mq_config = config.get("MQ")
        if not self.mq_config:
            raise ValueError(
                "Missing MQ configuration, please set it in ~/.config/neon/neon.yaml"
            )
        NeonAIClient.__init__(self, self.mq_config)
        self.router = APIRouter()
        self._await_response = Event()
        self._response = None
        self._transcribed = None
        self._current_tts = dict()
        self._profiles: Dict[str, dict] = dict()
        self._audio_path = join(
            xdg_data_home(), "iris", "stt"
        )  # TODO: Clear periodically, or have persistent storage
        if not isdir(self._audio_path):
            makedirs(self._audio_path)
        self.default_lang = lang or self.config.get("default_lang", "")
        LOG.name = "iris"
        LOG.init(self.config.get("logs"))
        # OpenWW
        # TODO: Allow for arbitrary models, or pre-existing OpenWW models
        self.oww_model = Model(
            wakeword_models=["neon_iris/wakeword_models/hey_neon/hey_neon_high.tflite"],
            inference_framework="tflite",
        )
        # FastAPI
        self.templates = Jinja2Templates(directory="neon_iris/templates")
        self.build_routes()

    def get_lang(self, session_id: str):
        """Get the language for a session."""
        if session_id and session_id in self._profiles:
            return self._profiles[session_id]["speech"]["stt_language"]
        return self.user_config["speech"]["stt_language"] or self.default_lang

    def handle_api_response(self, message: Message):
        """
        Catch-all handler for `.response` messages routed to this client that
        are not explicitly handled (i.e. get_stt, get_tts)
        @param message: Response message to something emitted by this client
        """
        LOG.debug(f"Got {message.msg_type}: {message.data}")
        if message.msg_type == "neon.audio_input.response":
            self._transcribed = message.data.get("transcripts", [""])[0]

    def handle_klat_response(self, message: Message):
        """
        Handle a valid response from Neon. This includes text and base64-encoded
        audio in all requested languages.
        @param message: Neon response message
        """
        LOG.debug(f"gradio context={message.context['gradio']}")
        resp_data = message.data["responses"]
        sentences = []
        session = message.context["gradio"]["session"]
        for _, response in resp_data.items():  # lang, response
            sentences.append(response.get("sentence"))
            if response.get("audio"):
                for _, data in response["audio"].items():
                    self._current_tts[session] = data
        self._response = "\n".join(sentences)
        self._await_response.set()

    def send_audio( # pylint: disable=arguments-renamed
        self,
        audio_b64_string: str,
        lang: str = "en-us",
        username: Optional[str] = None,
        user_profiles: Optional[list] = None,
        context: Optional[dict] = None,
    ):
        """
        Optionally override this to queue audio inputs or do any pre-parsing
        :param audio_file: path to audio file to send to speech module
        :param lang: language code associated with request
        :param username: username associated with request
        :param user_profiles: user profiles expecting a response
        :param context: Optional dict context to add to emitted message
        """
        audio_path = decode_base64_string_to_file(
            audio_b64_string,
            join(f"{self._audio_path}/{time()}.wav"),
        )
        self._send_audio(
            audio_file=audio_path,
            lang=lang,
            username=username,
            user_profiles=user_profiles,
            context=context,
        )

    @property
    def supported_languages(self) -> Sequence[str]:
        """
        Get a list of supported languages from configuration
        @returns: list of BCP-47 language codes
        """
        languages = self.config.get("languages")
        if languages is None:
            return [self.default_lang]
        if not isinstance(languages, list):
            raise TypeError("Expected a list of languages in the configuration")
        return languages

    def _start_session(self):
        sid = uuid4().hex
        self._current_tts[sid] = None
        self._profiles[sid] = self.user_config
        self._profiles[sid]["user"]["username"] = sid
        return sid

    def build_routes(self):
        """Build the FastAPI routes."""

        @self.router.get("/")
        async def read_root(request: Request):
            """Render the Neon AI Web UI and Voice Satellite."""
            description = self.config.get("webui_description", "Chat With Neon")
            title = self.config.get("webui_title", "Neon AI")
            placeholder = self.config.get("webui_input_placeholder", "Ask me something")
            ws_url = self.config.get("webui_ws_url", "ws://localhost:8000/ws")

            context = {
                "request": request,
                "title": title,
                "description": description,
                "placeholder": placeholder,
                "ws_url": ws_url
            }
            return self.templates.TemplateResponse("index.html", context)

        @self.router.websocket("/ws")
        async def websocket_endpoint(websocket: WebSocket):
            """Handles websocket connections to OpenWakeWord, which runs as part of this service."""
            await websocket.accept()
            # Send loaded models
            await websocket.send_text(
                json.dumps({"loaded_models": list(self.oww_model.models.keys())})
            )
            sample_rate = None

            while True:
                message = await websocket.receive()

                if message["type"] == "websocket.disconnect":
                    break

                if message["type"] == "websocket.receive":
                    if "text" in message:
                        # Process text message
                        sample_rate = int(message["text"])
                    elif "bytes" in message:
                        # Process bytes message
                        audio_bytes = message["bytes"]

                        # Add extra bytes of silence if needed
                        if len(audio_bytes) % 2 == 1:
                            audio_bytes += b"\x00"

                        # Convert audio to correct format and sample rate
                        audio_data = np.frombuffer(audio_bytes, dtype=np.int16)
                        if sample_rate and sample_rate != 16000:
                            audio_data = resampy.resample(
                                audio_data, sample_rate, 16000
                            )

                        # Get openWakeWord predictions and send to browser client
                        predictions = self.oww_model.predict(audio_data)

                        activations = [
                            key for key, value in predictions.items() if value >= 0.5
                        ]

                        if activations:
                            await websocket.send_text(
                                json.dumps({"activations": activations})
                            )

        @self.router.post("/user_input")
        async def on_user_input_worker(
            req: UserInput,
        ):
            """
            Callback to handle textual user input
            @param utterance: String utterance submitted by the user
            @returns: Session ID, audio input, audio output
            """
            utterance = req.utterance or ""
            audio_input = req.audio_input or ""
            session_id = req.session_id or "websat0000"

            chat_history = []
            input_time = time()
            LOG.debug("Input received")
            if not self._profiles.get("session_id"):
                self._profiles[session_id] = {
                    "speech": {"stt_language": self.default_lang}
                }
                self._current_tts[session_id] = None
            if not self._await_response.wait(30):
                LOG.error("Previous response not completed after 30 seconds")
            in_queue = time() - input_time
            self._await_response.clear()
            self._response = None
            self._transcribed = None
            lang = self.get_lang(session_id)
            if utterance:
                LOG.info(f"Sending utterance: {utterance} with lang: {lang}")
                self.send_utterance(
                    utterance,
                    lang or "en-us",
                    username=session_id,
                    user_profiles=[self._profiles[session_id]],
                    context={
                        "gradio": {"session": session_id},
                        "timing": {"wait_in_queue": in_queue, "gradio_sent": time()},
                    },
                )
            else:
                LOG.info(f"Sending audio with length of {len(audio_input)} with lang: {lang}")
                self.send_audio(
                    audio_input,
                    lang or "en-us",
                    username=session_id,
                    user_profiles=[self._profiles[session_id]],
                    context={
                        "gradio": {"session": session_id},
                        "timing": {"wait_in_queue": in_queue, "gradio_sent": time()},
                    },
                )
                chat_history.append(((audio_input, None), None))
            if not self._await_response.wait(30):
                LOG.error("No response received after 30s")
                self._await_response.set()
            self._response = self._response or "ERROR"
            LOG.info(f"Got response={self._response}")
            if utterance:
                chat_history.append((utterance, self._response))
            elif isinstance(self._transcribed, str):
                LOG.info(f"Got transcript: {self._transcribed}")
                chat_history.append((self._transcribed, self._response))
                utterance = self._transcribed
            resp = UserInputResponse(
                **{
                    "utterance": utterance,
                    "audio_output": self._current_tts[session_id],
                    "session_id": session_id,
                    "transcription": self._response,
                }
            )
            return resp


app = FastAPI()
neon_client = WebSatNeonClient()
app.mount(
    "/static",
    StaticFiles(directory="neon_iris/static"),
    name="Neon Web Voice Satellite",
)
app.include_router(neon_client.router)


if __name__ == "__main__":
    import uvicorn

    uvicorn.run(app, host="0.0.0.0", port=8000)
